(* $Id: atd_ast.ml 46617 2010-08-11 18:13:30Z martin $ *)

open Lexing


type loc = Lexing.position * Lexing.position

let dummy_loc = (Lexing.dummy_pos, Lexing.dummy_pos)

exception Atd_error of string

type full_module = module_head * module_body

and module_head = loc * annot

and module_body = module_item list

and annot = annot_section list

and annot_section = string * (loc * annot_field list)

and annot_field = string * (loc * string option)

and type_def = loc * (string * type_param * annot) * type_expr

and module_item =
    [ `Type of type_def ]

and type_param = string list

and type_expr = 
    [ `Sum of (loc * variant list * annot)
    | `Record of (loc * field list * annot)
    | `Tuple of (loc * cell list * annot)
    | `List of (loc * type_expr * annot)
    | `Option of (loc * type_expr * annot)
    | `Name of (loc * type_inst * annot)
    | `Tvar of (loc * string)
    ]
      (* `List and `Option are the 2 only predefined types with a type
	 parameter (and no special syntax). *)

and type_inst = loc * string * type_expr list

and variant =
    [ `Variant of (loc * (string * annot) * type_expr option)
    | `Inherit of (loc * type_expr) ]

and cell = loc * type_expr * annot

and field_kind =
    [ `Required
    | `Optional
    | `With_default
    ]

and field =
    [ `Field of (loc * (string * field_kind * annot) * type_expr)
    | `Inherit of (loc * type_expr) ]


let loc_of_type_expr = function
    `Sum (loc, _, _)
  | `Record (loc, _, _)
  | `Tuple (loc, _, _)
  | `List (loc, _, _)
  | `Option (loc, _, _)
  | `Name (loc, _, _)
  | `Tvar (loc, _) -> loc

let string_of_loc (pos1, pos2) =
  let line1 = pos1.pos_lnum
  and start1 = pos1.pos_bol in
  Printf.sprintf "File %S, line %i, characters %i-%i"
    pos1.pos_fname line1
    (pos1.pos_cnum - start1)
    (pos2.pos_cnum - start1)

let error s = raise (Atd_error s)


let annot_of_type_expr = function
    `Sum (_, _, an)
  | `Record (_, _, an)
  | `Tuple (_, _, an)
  | `List (_, _, an)
  | `Option (_, _, an)
  | `Name (_, _, an) -> an
  | `Tvar (_, _) -> []

 

module Type_names = Set.Make (String)

let union l =
  List.fold_left Type_names.union Type_names.empty l

let extract_type_names ?(ignorable = []) x =
  let ign s = List.mem s ignorable in

  let add s set =
    if ign s then set
    else Type_names.add s set
  in

  let rec x_expr (x : type_expr) : Type_names.t =
    match x with
	`Sum (loc, variant_list, annot) -> 
	  union (List.map x_variant variant_list)
	    
      | `Record (loc, field_list, annot) ->
	  union (List.map x_field field_list)
	    
      | `Tuple (loc, l, annot) ->
	  union (List.map (fun (loc, x, _) -> x_expr x) l)
	    
      | `List (loc, type_expr, annot) ->
	  x_expr type_expr
	    
      | `Option (loc, type_expr, annot) ->
	  x_expr type_expr
	    
      | `Name (loc, (loc2, name, type_expr_list), annot) ->
	  add name (union (List.map x_expr type_expr_list))
	    
      | `Tvar (loc, string) ->
	  Type_names.empty
	    
  and x_variant = function
      `Variant (loc, _, Some type_expr) -> x_expr type_expr
    | `Variant _ -> Type_names.empty
    | `Inherit (loc, type_expr) -> x_expr type_expr
	
  and x_field = function
      `Field (loc, _, type_expr) -> x_expr type_expr
    | `Inherit (loc, type_expr) -> x_expr type_expr

  in

  Type_names.elements (x_expr x)


let map_annot f = function
    `Sum (loc, vl, a) ->  `Sum (loc, vl, f a)
  | `Record (loc, fl, a) -> `Record (loc, fl, f a)
  | `Tuple (loc, tl, a) -> `Tuple (loc, tl, f a)
  | `List (loc, t, a) -> `List (loc, t, f a)
  | `Option (loc, t, a) -> `Option (loc, t, f a)
  | `Tvar _ as x -> x
  | `Name (loc, (loc2, name, args), a) ->
      `Name (loc, (loc2, name, args), f a)
